"""Client for the deepstream.io realtime web server"""
from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from deepstreampy.client import connect


__all__ = ["connect"]


version = "0.1.1"
version_info = (0, 1, 1)



from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from deepstreampy.constants import connection_state

from deepstreampy import client
from tests.util import msg

from tornado import testing
import sys

if sys.version_info[0] < 3:
    import mock
else:
    from unittest import mock

URL = "ws://localhost:7777/deepstream"


class PresenceHandlerTest(testing.AsyncTestCase):

    def setUp(self):
        super(PresenceHandlerTest, self).setUp()
        options = {"subscriptionTimeout": 0.005}
        self.client = client.Client(URL, **options)
        self.handler = mock.Mock()
        self.handler.stream.closed = mock.Mock(return_value=False)
        self.client._connection._state = connection_state.OPEN
        self.client._connection._websocket_handler = self.handler
        self.connection = self.client._connection
        self.io_loop = self.connection._io_loop
        self.client_errors = []

    def _stop_io_loop(self, *args, **kwargs):
        self.stop()

    def test_timeout(self):
        callback = mock.Mock(side_effect=self._stop_io_loop)
        error_callback = mock.Mock(side_effect=self._stop_io_loop)
        self.client.on('error', error_callback)

        # Send subscribe message
        self.client.presence.subscribe(callback)
        self.handler.write_message.assert_called_with(msg('U|S|U+'))

        # Emit timeout error
        self.wait()
        error_callback.assert_called_with(
            'No ACK message received in time for S', 'ACK_TIMEOUT', 'U')
        error_callback.reset_mock()

        # receive ack message for subscribe
        self.client.presence.handle({'topic': 'U',
                                      'action': 'A',
                                      'data': ['S']})
        self.wait()
        error_callback.assert_called_with('', 'UNSOLICITED_MESSAGE', 'U')
        error_callback.reset_mock()

        # Client logs in
        self.client.presence.handle({'topic': 'U',
                                      'action': 'PNJ',
                                      'data': ['Homer']})
        callback.assert_called_with('Homer', True)
        self.wait()

        # Client logs out
        self.client.presence.handle({'topic': 'U',
                                      'action': 'PNL',
                                      'data': ['Marge']})
        callback.assert_called_with('Marge', False)
        callback.reset_mock()
        self.wait()

        # Query for clients
        f = self.client.presence.get_all()
        f.add_done_callback(callback)
        self.handler.write_message.assert_called_with(msg('U|Q|Q+'))

        # Receive data for query
        self.client.presence.handle({'topic': 'U',
                                      'action': 'Q',
                                      'data': ['Marge', 'Homer', 'Bart']})

        self.wait()

        callback.assert_called_with(f)
        self.assertEquals(f.result(), ['Marge', 'Homer', 'Bart'])


        # Unsubscribe to client logins
        self.client.presence.unsubscribe(callback)
        self.handler.write_message.assert_called_with(msg('U|US+'))

        # emit ack timeout for unsubscribe
        self.wait()
        error_callback.assert_called_with(
            'No ACK message received in time for US', 'ACK_TIMEOUT', 'U')
        error_callback.reset_mock()

        # receive ack for unsubscrube
        self.client.presence.handle({'topic': 'U',
                                      'action': 'A',
                                      'data': ['US']})
        self.wait()
        error_callback.assert_called_with('', 'UNSOLICITED_MESSAGE', 'U')

        callback.reset_mock()
        error_callback.reset_mock()

        # not notified of future actions
        self.client.presence.handle({'topic': 'U',
                                      'action': 'PNJ',
                                      'data': ['Homer']})
        callback.assert_not_called()

        self.client.presence.handle({'topic': 'U',
                                      'action': 'PNL',
                                      'data': ['Homer']})
        callback.assert_not_called()
        self.client.presence.handle({'topic': 'U',
                                      'action': 'Q',
                                      'data': ['Marge', 'Homer', 'Bart']})
        callback.assert_not_called()



from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from deepstreampy.constants import topic as topic_constants
from deepstreampy.constants import actions as action_constants
from deepstreampy.constants import event as event_constants
from deepstreampy.utils import AckTimeoutRegistry
from deepstreampy.utils import ResubscribeNotifier

from tornado import concurrent
from tornado import gen

import json


class PresenceHandler(object):
    def __init__(self, connection, client, **options):
        self._options = options
        self._connection = connection
        self._client = client
        self._callbacks = {}
        self._query_future = None
        subscription_timeout = options.get("subscriptionTimeout", 15)
        self._ack_timeout_registry = AckTimeoutRegistry(
            client, topic_constants.PRESENCE, subscription_timeout)
        self._resubscribe_notifier = ResubscribeNotifier(
            client, self._resubscribe)

    @gen.coroutine
    def get_all(self):
        self._query_future = concurrent.Future()
        self._connection.send_message(topic_constants.PRESENCE,
                                             action_constants.QUERY,
                                             [action_constants.QUERY])
        result = yield self._query_future
        raise gen.Return(result)

    @gen.coroutine
    def get(self, users):
        self._query_future = concurrent.Future()
        self._connection.send_message(topic_constants.PRESENCE,
                                             action_constants.QUERY, users)
        result = yield self._query_future
        raise gen.Return(result)

    def subscribe(self, callback, users=None):
        if users is None:
            self._callbacks[topic_constants.PRESENCE] = callback
            users_str = topic_constants.PRESENCE
        else:
            for user in users:
                self._callbacks[user] = callback
            users_str = ",".join(users)

        self._ack_timeout_registry.add(action_constants.SUBSCRIBE, users_str)

        return self._connection.send_message(
            topic_constants.PRESENCE, action_constants.SUBSCRIBE, users_str)

    def unsubscribe(self, callback, users=None):
        users_str = ""

        if users is None:
            del self._callbacks[topic_constants.PRESENCE]
        else:
            for user in users:
                del self._callbacks[user]
            users_str = ",".join(users)

        self._ack_timeout_registry.add(action_constants.UNSUBSCRIBE, users_str)

        return self._connection.send_message(
            topic_constants.PRESENCE, action_constants.UNSUBSCRIBE, users_str)

    def handle(self, message):
        action = message['action']
        data = message['data']
        if (action == action_constants.ERROR
                and data[0] == event_constants.MESSAGE_DENIED):
            self._ack_timeout_registry.remove(topic_constants.PRESENCE,
                                              data[1])
            message['processedError'] = True
            self._client._on_error(topic_constants.PRESENCE,
                                   event_constants.MESSAGE_DENIED, data[1])
        elif action == action_constants.ACK:
            self._ack_timeout_registry.clear(message)
        elif action == action_constants.PRESENCE_JOIN:
            user = data[0]
            if user in self._callbacks.keys():
                self._callbacks[user](user, True)
            if topic_constants.PRESENCE in self._callbacks:
                self._callbacks[topic_constants.PRESENCE](user, True)
        elif action == action_constants.PRESENCE_LEAVE:
            user = data[0]
            if user in self._callbacks.keys():
                self._callbacks[user](user, False)
            if topic_constants.PRESENCE in self._callbacks:
                self._callbacks[topic_constants.PRESENCE](user, False)
        elif action == action_constants.QUERY:
            parsed = self._parse_query_response(data)
            if self._query_future:
                self._query_future.set_result(parsed)
                self._query_future = None
        else:
            self._client._on_error(topic_constants.PRESENCE,
                                   event_constants.UNSOLICITED_MESSAGE, action)

    def _parse_query_response(self, response):
        if response and response[0].isdigit():
            data = json.loads(response[1])
            return data
        return response

    def _resubscribe(self):
        if self._callbacks:
            self._connection.send_message(topic_constants.PRESENCE,
                                          action_constants.SUBSCRIBE,
                                          [action_constants.SUBSCRIBE])



"""
Tests for connecting to a client, logging in, state changes, and whether the
respective callbacks are made, and events triggered.
"""
from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from deepstreampy.message import connection
from deepstreampy import constants

from tornado import testing
import unittest

import sys
import errno

if sys.version_info[0] < 3:
    import mock
else:
    from unittest import mock

URL = "ws://localhost:7777/deepstream"

test_server_exceptions = []


class ConnectionTest(unittest.TestCase):

    def setUp(self):
        super(ConnectionTest, self).setUp()
        self.client = mock.Mock()
        self.handler = mock.Mock()

    def _get_connection_state_changes(self):
        count = 0
        for call_args in self.client.emit.call_args_list:
            if call_args[0][0] == constants.event.CONNECTION_STATE_CHANGED:
                count += 1
        return count

    def _get_sent_messages(self):
        for call_args in self.handler.write_message.call_args_list:
            yield call_args[0]

    def _get_last_sent_message(self):
        return self.handler.write_message.call_args[0][0]

    def test_connects(self):
        conn = connection.Connection(self.client, URL)
        assert conn.state == constants.connection_state.CLOSED
        self.assertEqual(self._get_connection_state_changes(), 0)
        connect_future = mock.Mock()
        connect_future_config = {'exception.return_value': None,
                                 'result.return_value': self.handler}
        connect_future.configure_mock(**connect_future_config)
        connect_future.exception.return_value = None
        connect_future.get_result.return_value = self.handler

        conn.connect()
        conn._on_open(connect_future)
        self.handler.stream.closed = mock.Mock(return_value=False)
        self.assertTrue(conn._websocket_handler is self.handler)
        self.assertEqual(conn.state,
                         constants.connection_state.AWAITING_CONNECTION)
        self.assertEqual(self._get_connection_state_changes(), 1)
        conn._on_data('C{0}A{1}'.format(chr(31), chr(30)))
        self.assertEqual(conn.state,
                         constants.connection_state.AWAITING_AUTHENTICATION)
        self.handler.write_message.assert_not_called()

        conn.authenticate({'user': 'Anon'})
        self.assertEqual(conn.state,
                         constants.connection_state.AUTHENTICATING)

        self.assertEqual(self._get_last_sent_message(),
                         "A{0}REQ{0}{{\"user\":\"Anon\"}}{1}".format(
                              chr(31), chr(30)).encode())
        self.assertEqual(self._get_connection_state_changes(), 3)

        conn._on_data('A{0}A{1}'.format(chr(31), chr(30)))
        self.assertEqual(conn.state,
                         constants.connection_state.OPEN)
        self.assertEqual(self._get_connection_state_changes(), 4)

        conn.send_message('R', 'S', ['test1'])
        self.handler.write_message.assert_called_with(
            'R{0}S{0}test1{1}'.format(chr(31), chr(30)).encode())

        conn.close()
        conn._on_close()

        self.assertEqual(conn.state,
                         constants.connection_state.CLOSED)
        self.assertEqual(self._get_connection_state_changes(), 5)

    def test_connect_error(self):
        conn = connection.Connection(self.client, URL)
        assert conn.state == constants.connection_state.CLOSED
        self.assertEqual(self._get_connection_state_changes(), 0)
        connect_future = mock.Mock()
        connect_future_config = {'exception.return_value': None,
                                 'result.return_value': self.handler}
        connect_future.configure_mock(**connect_future_config)
        connect_future.exception.return_value = IOError(
            (errno.ECONNREFUSED, "Connection refused"))
        connect_future.get_result.return_value = self.handler

        conn._on_open(connect_future)
        self.assertEqual(conn.state,
                         constants.connection_state.RECONNECTING)
        self.assertTrue(conn._reconnect_timeout is not None)
        conn._on_open(connect_future)
        conn._on_open(connect_future)
        conn._on_open(connect_future)
        self.assertEqual(conn.state,
                         constants.connection_state.ERROR)

    def test_too_many_auth_attempts(self):
        conn = connection.Connection(self.client, URL)
        connect_future = mock.Mock()
        connect_future_config = {'exception.return_value': None,
                                 'result.return_value': self.handler}
        connect_future.configure_mock(**connect_future_config)
        connect_future.exception.return_value = None
        connect_future.get_result.return_value = self.handler

        conn._on_open(connect_future)
        self.handler.stream.closed = mock.Mock(return_value=False)
        conn._on_data('C{0}A{1}'.format(chr(31), chr(30)))
        self.handler.write_message.assert_not_called()

        conn.authenticate({'user': 'Anon'})
        conn._on_data('A{0}E{1}'
                      .format(chr(31), chr(30)))
        conn.authenticate({'user': 'Anon'})
        conn._on_data('A{0}E{0}TOO_MANY_AUTH_ATTEMPTS{1}'
                      .format(chr(31), chr(30)))
        self.assertTrue(conn._too_many_auth_attempts)


class TestHeartbeat(testing.AsyncTestCase):

    def setUp(self):
        super(TestHeartbeat, self).setUp()
        client = mock.Mock()
        self.connection = connection.Connection(
            client, "ws://localhost2",
            heartbeatInterval=0.05)
        self.connection._io_loop = self.io_loop

        self.handler = mock.Mock()
        self.handler.stream.closed = mock.Mock(return_value=False)
        self.handler.close = mock.Mock(side_effect=self.connection._on_close)
        self._websocket_handler = self.handler

        future = mock.Mock()
        future.exception = mock.Mock(return_value=None)
        future.result = mock.Mock(return_value=self.handler)

        exc_future = mock.Mock()
        exc_future.exception = mock.Mock(return_value=None)
        exc_future.result = mock.Mock(return_value=self.handler)
        self.connection.connect = mock.Mock(
            return_value=future,
            side_effect=lambda: self.connection._on_open(exc_future))
        self.connection._on_open(future)
        self.connection._set_state(
            constants.connection_state.AWAITING_AUTHENTICATION)
        self.connection._on_data("C{0}A{1}".format(chr(31), chr(30)))

    def test_ping_pong(self):
        self.connection._on_data("C{0}PI{1}".format(chr(31), chr(30)))
        self.handler.write_message.assert_called_with(
            "C{0}PO{1}".format(chr(31), chr(30)).encode())

    @testing.gen_test
    def test_miss_one(self):
        yield testing.gen.sleep(0.075)
        self.handler.write_message.assert_not_called()
        self.assertEqual(self.connection.state,
                         constants.connection_state.AWAITING_AUTHENTICATION)

    @testing.gen_test
    def test_miss_two(self):
        yield testing.gen.sleep(3)
        self.assertEqual(self.connection.state,
                         constants.connection_state.ERROR)


if __name__ == '__main__':
    testing.unittest.main()



from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals


def remote_wins(record, remote_value, remote_version, callback):
    callback(None, remote_value)


def local_wins(record, remote_value, remote_version, callback):
    callback(None, record.get())



from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

CONNECTION = 'C'
AUTH = 'A'
ERROR = 'X'
EVENT = 'E'
RECORD = 'R'
RPC = 'P'
PRESENCE = 'U'
PRIVATE = 'PRIVATE/'



from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from deepstreampy.constants import connection_state

from deepstreampy import client
from tests.util import msg
import unittest
import sys

if sys.version_info[0] < 3:
    import mock
else:
    from unittest import mock

URL = "ws://localhost:7777/deepstream"


class EventsTest(unittest.TestCase):

    def setUp(self):
        super(EventsTest, self).setUp()

        self.client = client.Client(URL)
        self.handler = mock.Mock()
        self.handler.stream.closed = mock.Mock(return_value=False)
        self.client._connection._state = connection_state.OPEN
        self.client._connection._websocket_handler = self.handler
        self.connection = self.client._connection

        self.event_callback = mock.Mock()
        self.error_callback = mock.Mock()

    def test_handler(self):
        self.handler.write_message.assert_not_called()
        self.client.event.emit('myEvent', 6)
        self.handler.write_message.assert_called_with(msg('E|EVT|myEvent|N6+'))

        self.client.event.subscribe('myEvent', self.event_callback)
        self.handler.write_message.assert_called_with(msg('E|S|myEvent+'))

        self.client.on('error', self.error_callback)
        self.event_callback.assert_not_called()
        self.client.event.handle({'topic': 'EVENT',
                                  'action': 'EVT',
                                  'data': ['myEvent', 'N23']})
        self.event_callback.assert_called_with(23)

        self.client.event.handle({'topic': 'EVENT',
                                  'action': 'EVT',
                                  'data': ['myEvent']})
        self.event_callback.assert_called_with()

        self.client.event.handle({'topic': 'EVENT',
                                  'action': 'EVT',
                                  'data': ['myEvent', 'notTypes']})
        self.error_callback.assert_called_with('UNKNOWN_TYPE (notTypes)',
                                               'MESSAGE_PARSE_ERROR',
                                               'X')
        self.event_callback.reset_mock()
        self.client.event.unsubscribe('myEvent', self.event_callback)
        self.client.event.emit('myEvent', 11)
        self.event_callback.assert_not_called()

        self.client.event.handle({'topic': 'EVENT',
                                  'action': 'L',
                                  'data': ['myEvent']})
        self.error_callback.assert_called_with('myEvent',
                                               'UNSOLICITED_MESSAGE',
                                               'E')

    def test_accept(self):
        def listen_callback(data, is_subscribed, response):
            response.accept()

        self.client.event.listen('a/.*', listen_callback)
        self.client.event.handle({'topic': 'E',
                                  'action': 'SP',
                                  'data': ['a/.*', 'a/1']})

        self.handler.write_message.assert_called_with(msg('E|LA|a/.*|a/1+'))

    def test_reject(self):
        def listen_callback(data, is_subscribed, response):
            response.reject()

        self.client.event.listen('b/.*', listen_callback)
        self.client.event.handle({'topic': 'E',
                                  'action': 'SP',
                                  'data': ['b/.*', 'b/1']})

        self.handler.write_message.assert_called_with(msg('E|LR|b/.*|b/1+'))

    def test_accept_and_discard(self):
        def listen_callback(data, is_subscribed, response=None):
            if is_subscribed:
                response.accept()

                self.client.event.handle({'topic': 'E',
                                          'action': 'SR',
                                          'data': ['b/.*', 'b/2']})

        self.client.event.listen('b/.*', listen_callback)
        self.client.event.handle({'topic': 'E',
                                  'action': 'SP',
                                  'data': ['b/.*', 'b/2']})

        self.handler.write_message.assert_called_with(msg('E|LA|b/.*|b/2+'))

    def test_accept_unlisten(self):

        def listen_callback(data, is_subscribed, response):
            response.accept()

        self.client.event.listen('a/.*', listen_callback)
        self.client.event.handle({'topic': 'E',
                                  'action': 'SP',
                                  'data': ['a/.*', 'a/1']})

        self.handler.write_message.assert_called_with(msg('E|LA|a/.*|a/1+'))

        self.client.event.unlisten('a/.*')
        self.handler.write_message.assert_called_with(msg('E|UL|a/.*+'))

        self.handler.reset_mock()
        self.client.event.handle({'topic': 'E',
                                  'action': 'A',
                                  'data': ['UL', 'a/.*']})
        self.client.on('error', self.error_callback)
        self.client.event.handle({'topic': 'E',
                                  'action': 'SP',
                                  'data': ['a/.*', 'a/1']})
        self.error_callback.assert_called_with('a/.*',
                                               'UNSOLICITED_MESSAGE',
                                               'E')
        self.handler.write_message.assert_not_called()

        self.client.event.unlisten('a/.*')
        self.error_callback.assert_called_with('a/.*',
                                               'NOT_LISTENING',
                                               'X')

    def test_existing_listener(self):
        self.client.on('error', self.error_callback)

        def listen_callback(data, is_subscribed, response):
            response.accept()

        self.client.event.listen('b/.*', listen_callback)
        self.client.event.listen('b/.*', listen_callback)



from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from deepstreampy import client
from deepstreampy import rpc
from deepstreampy import constants
from tests.util import msg

from tornado import testing
import functools
import sys

if sys.version_info[0] < 3:
    import mock
else:
    from unittest import mock

URL = "ws://localhost:7777/deepstream"


class RPCHandlerTest(testing.AsyncTestCase):

    def setUp(self):
        super(RPCHandlerTest, self).setUp()
        options = {'rpcResponseTimeout': 0.01,
                   'rpcAckTimeout': 0.01,
                   'subscriptionTimeout': 0.01}
        self.client = client.Client(URL, **options)
        self.handler = mock.Mock()
        self.handler.stream.closed = mock.Mock(return_value=False)
        self.client._connection._state = constants.connection_state.OPEN
        self.client._connection._websocket_handler = self.handler
        self.connection = self.client._connection
        self.io_loop = self.connection._io_loop
        self.rpc_calls = 0
        self.client_errors = []
        self.client.on('error', self._error_callback)
        get_uid_patcher = mock.patch(
            'deepstreampy.utils.get_uid', return_value='1')
        get_uid_patcher.start()
        self.addCleanup(get_uid_patcher.stop)

    def _error_callback(self, *args):
        self.client_errors.append(args)
        self.stop()

    def _add_two_callback(self, data, response):
        self.rpc_calls += 1
        result = data['numA'] + data['numB']
        if 'sync' in data and data['sync']:
            response.send(result)
            self.stop()
        else:
            self.connection._io_loop.call_later(
                0.5, functools.partial(response.send, result))
            self.connection._io_loop.call_later(0.6, self.stop)

    def test_validate_params(self):
        with self.assertRaises(ValueError):
            self.client.rpc.provide('', lambda: None)

        with self.assertRaises(TypeError):
            self.client.rpc.provide(123, lambda: None)

        with self.assertRaises(ValueError):
            self.client.rpc.provide(None, lambda: None)

        with self.assertRaises(ValueError):
            self.client.rpc.provide('name', None)

        with self.assertRaises(TypeError):
            self.client.rpc.provide('name', 123)

    def test_handle_rpc_providers(self):
        # RPCHandler is created
        rpchandler = self.client.rpc
        self.assertTrue(isinstance(self.client.rpc, rpc.RPCHandler))

        # Register a provider for addTwo RPC
        rpchandler.provide('addTwo', self._add_two_callback)
        self.handler.write_message.assert_called_once_with(msg('P|S|addTwo+'))
        self.assertEquals(self.rpc_calls, 0)

        # Timeout error emitted if no ack message received on time
        self.wait()
        expected_error = ('No ACK message received in time for addTwo',
                          'ACK_TIMEOUT',
                          'P')
        self.assertIn(expected_error, self.client_errors)

        # Reply to a sync RPC request
        rpchandler.handle({'topic': 'RPC',
                           'action': 'REQ',
                           'data': ['addTwo',
                                    '678',
                                    'O{"numA":2,"numB":3,"sync":true}']})
        self.wait()
        self.handler.write_message.assert_called_with(
            msg('P|RES|addTwo|678|N5+'))

        # Reply to an async RPC request
        rpchandler.handle({'topic': 'RPC',
                           'action': 'REQ',
                           'data': ['addTwo', '123', 'O{"numA":7,"numB":3}']})

        self.connection._io_loop.call_later(0.1, self.stop)
        self.wait()
        self.handler.write_message.assert_called_with(
            msg('P|A|REQ|addTwo|123+'))

        self.wait()
        self.handler.write_message.assert_called_with(
            msg('P|RES|addTwo|123|N10+'))

        # Send rejection if no provider exists
        rpchandler.handle({'topic': 'RPC',
                           'action': 'REQ',
                           'data': ['doesNotExist', '432',
                                    'O{"numA":7,"numB":3}']})
        self.handler.write_message.assert_called_with(
            msg('P|REJ|doesNotExist|432+'))

        # Deregister provider for the addTwo RPC
        rpchandler.unprovide('addTwo')
        self.handler.write_message.assert_called_with(msg('P|US|addTwo+'))

        # Timeout emitted after no ACK message received for the unprovide
        self.client_errors = []
        self.connection._io_loop.call_later(3.5, self.stop)
        self.wait()
        expected_error = ('No ACK message received in time for addTwo',
                          'ACK_TIMEOUT',
                          'P')
        self.assertIn(expected_error, self.client_errors)

        # Reject call to deregistered provider
        rpchandler.handle({
            'topc': 'RPC',
            'action': 'REQ',
            'data': ['addTwo', '434', 'O{"numA":2,"numB":7, "sync": true}']})

        self.handler.write_message.assert_called_with(msg('P|REJ|addTwo|434+'))

    @testing.gen_test
    def test_make_rpcs(self):
        # RPCHandler is created
        rpchandler = self.client.rpc
        self.assertTrue(isinstance(self.client.rpc, rpc.RPCHandler))

        # Make a successful RPC for addTwo
        f = rpchandler.make('addTwo', {'numA': 3, 'numB': 8})
        self.assertTrue(self.handler.write_message.call_args[0][0] in
                        (msg('P|REQ|addTwo|1|O{"numA":3,"numB":8}+'),
                         msg('P|REQ|addTwo|1|O{"numB":8,"numA":3}+')))

        rpchandler.handle({'topic': 'RPC',
                           'action': 'RES',
                           'data': ['addTwo', u'1', 'N11']})

        result = yield f
        self.assertEquals(result, 11)

        # Make RPC for addTwo but receive an error
        self.assertTrue(self.handler.write_message.call_args[0][0] in
                        (msg('P|REQ|addTwo|1|O{"numA":3,"numB":8}+'),
                         msg('P|REQ|addTwo|1|O{"numB":8,"numA":3}+')))

        f = rpchandler.make('addTwo', {'numA': 3, 'numB': 8})
        rpchandler.handle({'topic': 'RPC',
                           'action': 'E',
                           'data': ['NO_PROVIDER', 'addTwo', '1']})

        with self.assertRaises(rpc.RPCException) as ectx:
            yield f

        self.assertEquals(str(ectx.exception), 'NO_PROVIDER')

        # Make RPC for addTwo but receive no ack in time
        f = rpchandler.make('addTwo', {'numA': 3, 'numB': 8})
        self.assertTrue(self.handler.write_message.call_args[0][0] in
                        (msg('P|REQ|addTwo|1|O{"numA":3,"numB":8}+'),
                         msg('P|REQ|addTwo|1|O{"numB":8,"numA":3}+')))

        with self.assertRaises(rpc.RPCException) as ectx:
            yield f

        self.assertEquals(str(ectx.exception), 'ACK_TIMEOUT')


class RPCResponseTest(testing.AsyncTestCase):

    def setUp(self):
        super(RPCResponseTest, self).setUp()
        options = {'rpcResponseTimeout': 0.01, 'rpcAckTimeout': 0.01}
        self.client = client.Client(URL, **options)
        self.handler = mock.Mock()
        self.handler.stream.closed = mock.Mock(return_value=False)
        self.client._connection._state = constants.connection_state.OPEN
        self.client._connection._websocket_handler = self.handler
        self.connection = self.client._connection
        self.connection._io_loop = self.io_loop
        self.client_errors = []
        self.client.on('error', self._error_callback)
        get_uid_patcher = mock.patch(
            'deepstreampy.utils.get_uid', return_value='1')
        get_uid_patcher.start()
        self.addCleanup(get_uid_patcher.stop)

    def _error_callback(self, *args):
        self.client_errors.append(args)
        self.stop()

    def test_send(self):
        # Send auto ACK
        response = rpc.RPCResponse(self.connection, 'addTwo', '123')
        self.io_loop.call_later(1, self.stop)
        self.wait()
        self.handler.write_message.assert_called_with(
            msg('P|A|REQ|addTwo|123+'))

        # Send response
        response.send(14)
        self.handler.write_message.assert_called_with(
            msg('P|RES|addTwo|123|N14+'))

    @testing.gen_test
    def test_send_no_autoack(self):
        # Create response but disable autoack
        response = rpc.RPCResponse(self.connection, 'addTwo', '123')
        response.auto_ack = False
        self.handler.write_message.assert_not_called()

        # Send the ack
        response.ack()
        self.handler.write_message.assert_called_with(
            msg('P|A|REQ|addTwo|123+'))

        # Do not send multiple ack messages
        self.handler.reset_mock()
        response.ack()
        self.handler.write_message.assert_not_called()

    def test_reject(self):
        response = rpc.RPCResponse(self.connection, 'addTwo', '123')
        response.reject()
        self.handler.write_message.assert_called_with(
            msg('P|REJ|addTwo|123+'))
        self.assertRaises(ValueError, functools.partial(response.send, 'abc'))

    def test_error(self):
        response = rpc.RPCResponse(self.connection, 'addTwo', '123')
        response.error('Error message')
        self.handler.write_message.assert_called_with(
            msg('P|E|Error message|addTwo|123+'))
        self.assertRaises(ValueError, functools.partial(response.send, 'abc'))



from __future__ import absolute_import, division, print_function, with_statement
from __future__ import unicode_literals

from deepstreampy.constants import actions as action_constants
from deepstreampy.constants import event as event_constants
from deepstreampy.constants import connection_state

from pyee import EventEmitter
from tornado import concurrent

from functools import partial
from collections import namedtuple

import sys
import time
import random

num_types = ((int, long, float, complex)
             if sys.version_info < (3, ) else (int, float, complex))
str_types = (str, unicode) if sys.version_info < (3, ) else (str, )


class SingleNotifier(object):
    def __init__(self, client, connection, topic, action, timeout_duration):
        self._client = client
        self._connection = connection
        self._topic = topic
        self._action = action
        self._timeout_duration = timeout_duration

        self._requests = dict()

        self._resubscribe_notifier = ResubscribeNotifier(
            client, self._resend_requests)

    def has_request(self, name):
        return name in self._requests

    def request(self, name, callback):
        if name not in self._requests:
            self._requests[name] = []
            future = self._connection.send_message(self._topic, self._action,
                                                   [name])
        else:
            future = concurrent.Future()
            future.set_result()

        response_timeout = self._client.io_loop.call_later(
            self._timeout_duration, partial(self._on_response_timeout, name))
        self._requests[name].append({
            'timeout': response_timeout,
            'callback': callback
        })

        return future

    def receive(self, name, error, data):
        entries = self._requests[name]
        for entry in entries:
            self._client.io_loop.remove_timeout(entry['timeout'])
            entry['callback'](error, data)
        del self._requests[name]

    def _on_response_timeout(self, name):
        msg = ('No response received in time for '
               '{0}|{1}|{2}').format(self._topic, self._action, name)
        self._client._on_error(self._topic, event_constants.RESPONSE_TIMEOUT,
                               msg)

    def _resend_requests(self):
        for request in self._requests:
            self._connection.send_message(self._topic, self._action,
                                          [self._requests[request]])


CallbackResponse = namedtuple('CallbackResponse', 'accept reject')


class Listener(object):
    def __init__(self, listener_type, pattern, callback, options, client,
                 connection):
        self._type = listener_type
        self._callback = callback
        self._pattern = pattern
        self._options = options
        self._client = client
        self._connection = connection
        self._send_future = None

        subscription_timeout = options.get("subscriptionTimeout", 15)
        self._ack_timeout = connection._io_loop.call_later(
            subscription_timeout, self._on_ack_timeout)
        self._resubscribe_notifier = ResubscribeNotifier(
            client, self._send_listen)
        self._send_listen()
        self._destroy_pending = False

    def send_destroy(self):
        self._destroy_pending = True
        future = self._connection.send_message(
            self._type, action_constants.UNLISTEN, [self._pattern])
        self._resubscribe_notifier.destroy()
        return future

    def destroy(self):
        self._callback = None
        self._pattern = None
        self._client = None
        self._connection = None

    def accept(self, name):
        return self._connection.send_message(
            self._type, action_constants.LISTEN_ACCEPT, [self._pattern, name])

    def reject(self, name):
        return self._connection.send_message(
            self._type, action_constants.LISTEN_REJECT, [self._pattern, name])

    def _create_callback_response(self, message):
        return CallbackResponse(
            accept=partial(self.accept, message['data'][1]),
            reject=partial(self.reject, message['data'][1]))

    def _on_message(self, message):
        action = message['action']
        data = message['data']
        if action == action_constants.ACK:
            self._connection._io_loop.remove_timeout(self._ack_timeout)
        elif action == action_constants.SUBSCRIPTION_FOR_PATTERN_FOUND:
            # TODO: Show deprecated message
            self._callback(data[1], True,
                           self._create_callback_response(message))
        elif action == action_constants.SUBSCRIPTION_FOR_PATTERN_REMOVED:
            self._callback(data[1], False)
        else:
            is_found = (message['action'] ==
                        action_constants.SUBSCRIPTION_FOR_PATTERN_FOUND)
            self._callback(message['data'][1], is_found)

    def _send_listen(self):
        self._send_future = self._connection.send_message(
            self._type, action_constants.LISTEN, [self._pattern])

    def _on_ack_timeout(self):
        self._client._on_error(
            self._type, event_constants.ACK_TIMEOUT,
            ('No ACK message received in time for ' + self._pattern))

    @property
    def destroy_pending(self):
        return self._destroy_pending

    @property
    def send_future(self):
        return self._send_future


class ResubscribeNotifier(object):
    """
    Makes sure that all functionality is resubscribed on reconnect. Subscription
    is called when the connection drops - which seems counterintuitive, but in
    fact just means that the re-subscription message will be added to the queue
    of messages that need re-sending as soon as the connection is
    re-established.

    Resubscribe logic should only occur once per connection loss.
    """

    def __init__(self, client, resubscribe):
        """
        Args:
            client: The deepstream client
            resubscribe: callable to call to allow resubscribing
        """
        self._client = client
        self._resubscribe = resubscribe

        self._is_reconnecting = False
        self._client.on(event_constants.CONNECTION_STATE_CHANGED,
                        self._handle_connection_state_changes)

    def destroy(self):
        self._client.remove_listener(event_constants.CONNECTION_STATE_CHANGED,
                                     self._handle_connection_state_changes)
        self._client = None

    def _handle_connection_state_changes(self, state):
        if state == connection_state.RECONNECTING and not self._is_reconnecting:
            self._is_reconnecting = True
        elif state == connection_state.OPEN and self._is_reconnecting:
            self._is_reconnecting = False
            self._resubscribe()


class AckTimeoutRegistry(EventEmitter):
    def __init__(self, client, topic, timeout_duration):
        super(AckTimeoutRegistry, self).__init__()
        self._client = client
        self._topic = topic
        self._timeout_duration = timeout_duration
        self._register = {}

    def add(self, name, action=None):
        unique_name = (action or "") + name

        self.remove(unique_name, action)
        timeout = self._client.io_loop.call_later(self._timeout_duration,
                                                  partial(
                                                      self._on_timeout,
                                                      unique_name, name))
        self._register[unique_name] = timeout

    def remove(self, name, action=None):
        unique_name = (action or "") + name
        if unique_name in self._register:
            self.clear({'data': [action, name]})

    def clear(self, message):
        unique_name = "".join(message['data'][:2])

        if unique_name in self._register:
            timeout = self._register[unique_name]
            self._client.io_loop.remove_timeout(timeout)
        else:
            self._client._on_error(self._topic,
                                   event_constants.UNSOLICITED_MESSAGE,
                                   message.get('raw', ''))

    def _on_timeout(self, unique_name, name):
        del self._register[unique_name]
        msg = "No ACK message received in time for " + name
        self._client._on_error(self._topic, event_constants.ACK_TIMEOUT, msg)
        self.emit('timeout', name)


def _pad_list(l, index, value):
    l.extend([value] * (index - len(l)))


class Undefined(object):
    def __repr__(self):
        return 'Undefined'


Undefined = Undefined()


def itoa(num, radix):
    """
    Convert int to a sts representation in an arbitrary base, up to 36.
    """
    result = ""
    while num > 0:
        result = "0123456789abcdefghijklmnopqrstuvwxyz" [num % radix] + result
        num //= radix
    return result


def get_uid():
    timestamp = itoa(int(time.time() * 1000), 36)
    random_str = itoa(int(random.random() * 10000000000000000), 36)
    return "{0}-{1}".format(timestamp, random_str)
